from typing import Text, List, Optional, Dict, Callable, Any

import sqlalchemy
import sqlalchemy.exc
import sqlalchemy.orm
from alembic import op
from alembic.operations import BatchOperations
from sqlalchemy import engine_from_config, Column, MetaData, Table, schema, Sequence
from sqlalchemy.dialects.sqlite.base import SQLiteDialect
from sqlalchemy.engine import reflection


def table_has_column(table: Text, column: Text) -> bool:
    """Checks whether a table with a given column exists."""

    alembic_config = op.get_context().config
    engine = engine_from_config(
        alembic_config.get_section(alembic_config.config_ini_section),
        prefix="sqlalchemy.",
    )
    insp = reflection.Inspector.from_engine(engine)
    has_column = False
    try:
        for col in insp.get_columns(table):
            if column not in col["name"]:
                continue
            has_column = True
    except sqlalchemy.exc.NoSuchTableError:
        return False
    return has_column


def create_column(table: Text, column: Column) -> None:
    """Adds `column` to `table`."""

    with op.batch_alter_table(table) as batch_op:
        batch_op.add_column(column)


def drop_column(table: Text, column: Text) -> None:
    """Drops `column` from `table`."""

    with op.batch_alter_table(table) as batch_op:
        batch_op.drop_column(column)


def create_primary_key(
    batch_op: BatchOperations,
    table_name: Text,
    primary_key_columns: List[Text],
    primary_key_name: Optional[Text] = None,
) -> None:
    """Creates a new primary key and drops existing old ones."""

    old_primary_key_name = get_primary_key(table_name)
    drop_primary_key(old_primary_key_name, batch_op)
    if not primary_key_name:
        primary_key_name = old_primary_key_name

    batch_op.create_primary_key(primary_key_name, columns=primary_key_columns)


def get_primary_key(table_name: Text) -> Text:
    """Returns the name of the primary key in a table."""

    meta = MetaData(bind=op.get_bind())
    meta.reflect()
    table = Table(table_name, meta)

    return table.primary_key.name


def drop_primary_key(key_name: Text, batch_op: BatchOperations) -> None:
    """Drops the primary key of a table."""

    context = op.get_context()
    # in SQLite the constraint is unnamed and this does not work
    # (dropping the constraint is also not needed)
    if not isinstance(context.dialect, SQLiteDialect):
        batch_op.drop_constraint(key_name)


def rename_or_create_index(
    table_name: Text,
    old_index_name: Text,
    new_index_name: Text,
    indexed_columns: List[Text],
) -> None:
    """Rename an index in case it exists, create it otherwise."""

    meta = MetaData(bind=op.get_bind())
    meta.reflect()
    table = Table(table_name, meta)
    existing_indexes = [index.name for index in table.indexes]

    with op.batch_alter_table(table_name) as batch_op:
        # Databases might save indexes in a lower / upper case version
        matching_index_name = _find_matching_index(old_index_name, existing_indexes)

        if matching_index_name:
            batch_op.drop_index(matching_index_name)

        existing_new_index_name = _find_matching_index(new_index_name, existing_indexes)

        if not existing_new_index_name:
            batch_op.create_index(new_index_name, indexed_columns)


def _find_matching_index(
    index_name_to_find: Text, existing_indexes: List[Text]
) -> Optional[Text]:
    return next(
        (
            index_name
            for index_name in existing_indexes
            if index_name.lower() == index_name_to_find.lower()
        ),
        None,
    )


class ColumnTransformation:
    """Describes the modification which should be done to the column of a table."""

    def __init__(
        self,
        column_name: Text,
        new_column_args: Optional[List] = None,
        new_column_kwargs: Optional[Dict] = None,
        modify: Optional[Callable[[Any], Any]] = None,
    ) -> None:
        """Creates a transformation.

        Args:
            column_name: Current name of the column which should be changed.
            new_column_args: Args which are passed to `Column`.
            new_column_kwargs: Kwargs which are passed to `Column`.
            modify: Operation which should be applied when copying data from the old
                    column to the new column.
        """
        self.column_name = column_name
        self.new_column_args = new_column_args or []
        self.new_column_kwargs = new_column_kwargs or {}
        self.modify = modify


def modify_columns(table_name: Text, columns: List[ColumnTransformation]) -> None:
    """Modify multiple columns of a table."""

    bind = op.get_bind()
    session = sqlalchemy.orm.Session(bind=bind)

    with op.batch_alter_table(table_name) as batch_op:
        # Create temporary columns including modifications
        for transformation in columns:
            temporary_column_name = _get_temporary_column_name(
                transformation.column_name
            )
            batch_op.add_column(
                sqlalchemy.Column(
                    temporary_column_name,
                    *transformation.new_column_args,
                    **transformation.new_column_kwargs,
                )
            )

    with op.batch_alter_table(table_name) as batch_op:

        metadata = MetaData(bind=bind)
        metadata.reflect()
        table = Table(table_name, metadata, autoload=True)

        for transformation in columns:
            temporary_column_name = _get_temporary_column_name(
                transformation.column_name
            )

            # Use SQLAlchemy Core cause the ORM mappings can change over time
            entries = session.execute(table.select()).fetchall()
            for entry in entries:
                old_value = getattr(entry, transformation.column_name)
                if transformation.modify:
                    new_value = transformation.modify(old_value)
                else:
                    new_value = old_value
                update_query = (
                    sqlalchemy.update(table)
                    .values({temporary_column_name: new_value})
                    .where(table.c[transformation.column_name] == old_value)
                )
                session.execute(update_query)

            session.commit()

            # Delete old column
            batch_op.drop_column(transformation.column_name)
            # Rename temporary column to replace old one
            batch_op.alter_column(
                column_name=temporary_column_name,
                new_column_name=transformation.column_name,
            )


def _get_temporary_column_name(column_name: Text) -> Text:
    return column_name + "_tmp"


def dialect_supports_sequences() -> bool:
    return op._proxy.migration_context.dialect.supports_sequences


def create_sequence(table_name: Text) -> None:
    if dialect_supports_sequences():
        op.execute(schema.CreateSequence(Sequence(table_name + "_seq")))
